@@ -6,9 +6,12 @@
 #
 # Python DB-API 2.0 module for Firebird.
 ##############################################################################
+from __future__ import print_function
 import sys
+import warnings
 import xdrlib, time, datetime, decimal, struct
 import itertools
+from collections import Mapping
 from firebirdsql.fberrmsgs import messages
 from firebirdsql import (DatabaseError, InternalError, OperationalError,
     ProgrammingError, IntegrityError, DataError, NotSupportedError,)
@@ -17,37 +20,37 @@ from firebirdsql.utils import *
 from firebirdsql.wireprotocol import WireProtocol, INFO_SQL_SELECT_DESCRIBE_VARS
 from firebirdsql.socketstream import SocketStream
 from firebirdsql.xsqlvar import XSQLVAR, calc_blr, parse_select_items, parse_xsqlda
-
-try:
-    from collections import Mapping
-    HAS_MAPPING = True
-except ImportError:
-    # Python 2.5
-    from UserDict import DictMixin as Mapping
-    HAS_MAPPING = False
-
-__version__ = '0.9.0'
+__version__ = '0.9.1'
 apilevel = '2.0'
 threadsafety = 1
 paramstyle = 'qmark'
 
+DEBUG = False
+
+def DEBUG_OUTPUT(*argv):
+    if not DEBUG:
+        return
+    for s in argv:
+        print(s, end=' ', file=sys.stderr)
+    print(file=sys.stderr)
+
 transaction_parameter_block = (
-    # ISOLATION_LEVEL_READ_UNCOMMITTED
-    bytes([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_read_committed, isc_tpb_rec_version]),
+    # ISOLATION_LEVEL_READ_COMMITED_LEGACY
+    bs([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_read_committed, isc_tpb_no_rec_version]),
     # ISOLATION_LEVEL_READ_COMMITED
-    bytes([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_read_committed, isc_tpb_no_rec_version]),
+    bs([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_read_committed, isc_tpb_rec_version]),
     # ISOLATION_LEVEL_REPEATABLE_READ
-    bytes([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_concurrency]),
+    bs([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_concurrency]),
     # ISOLATION_LEVEL_SERIALIZABLE
-    bytes([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_consistency]),
-    # ISOLATION_LEVEL_READ_COMMITED_READ_ONLY
-    bytes([isc_tpb_version3, isc_tpb_read, isc_tpb_wait, isc_tpb_read_committed, isc_tpb_no_rec_version]),
+    bs([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_consistency]),
+    # ISOLATION_LEVEL_READ_COMMITED_RO
+    bs([isc_tpb_version3, isc_tpb_read, isc_tpb_wait, isc_tpb_read_committed,isc_tpb_rec_version]),
 )
 
-def Date(year, month, day):
-    return datetime.date(year, month, day)
-def Time(hour, minite, second):
-    return datetime.time(hour, minite, second)
+Date = datetime.date
+Time = datetime.time
+TimeDelta = datetime.timedelta
+Timestamp = datetime.datetime
 def DateFromTicks(ticks):
     return apply(Date,time.localtime(ticks)[:3])
 def TimeFromTicks(ticks):
@@ -55,7 +58,7 @@ def TimeFromTicks(ticks):
 def TimestampFromTicks(ticks):
     return apply(Timestamp,time.localtime(ticks)[:6])
 def Binary(b):
-    return b
+    return bs(b)
 
 class DBAPITypeObject:
     def __init__(self,*values):
@@ -74,6 +77,8 @@ else:
     BINARY = DBAPITypeObject(str)
 NUMBER = DBAPITypeObject(int, decimal.Decimal)
 DATETIME = DBAPITypeObject(datetime.datetime, datetime.date, datetime.time)
+DATE = DBAPITypeObject(datetime.date)
+TIME = DBAPITypeObject(datetime.time)
 ROWID = DBAPITypeObject()
 
 class Statement(object):
@@ -81,10 +86,11 @@ class Statement(object):
     statement handle and status (open/close)
     """
     def __init__(self, trans):
+        DEBUG_OUTPUT("Statement::__init__()")
         self.trans = trans
         self._allocate_stmt()
         self._is_open = False
-        self.trans.stmts.add(self)
+        self.stmt_type = None
 
     def _allocate_stmt(self):
         self.trans.connection._op_allocate_statement()
@@ -95,63 +101,54 @@ class Statement(object):
             self.handle = h
 
     def prepare(self, sql, explain_plan=False):
-        if self.handle == 0:
-            self._allocate_stmt()
-
+        DEBUG_OUTPUT("Statement::prepare()", self.handle)
         if explain_plan:
             self.trans.connection._op_prepare_statement(
                 self.handle, self.trans.trans_handle, sql,
-                option_items=bytes([isc_info_sql_get_plan]))
+                option_items=bs([isc_info_sql_get_plan]))
         else:
             self.trans.connection._op_prepare_statement(
                 self.handle, self.trans.trans_handle, sql)
             self.plan = None
 
         (h, oid, buf) = self.trans.connection._op_response()
-        if self.trans.connection.accept_type == ptype_lazy_send:
-            self.handle = h
 
         i = 0
         if byte_to_int(buf[i]) == isc_info_sql_get_plan:
             l = bytes_to_int(buf[i+1:i+3])
             self.plan = self.trans.connection.bytes_to_str(buf[i+3:i+3+l])
             i += 3 + l
-
         self.stmt_type, self.xsqlda = parse_xsqlda(buf[i:],
                                     self.trans.connection, self.handle)
 
     def open(self):
+        DEBUG_OUTPUT("Statement::open()")
         self._is_open = True
 
-    def _close(self):
-        self._is_open = False
-
     def close(self):
-        if not self._is_open or self.handle == 0 or self.handle == -1:
-            return
-        self.trans.connection._op_free_statement(self.handle, DSQL_close)
-        if self.trans.connection.accept_type != ptype_lazy_send:
-            (h, oid, buf) = self.trans.connection._op_response()
-        self._close()
+        DEBUG_OUTPUT("Statement::close()", self.handle)
+        if (self.stmt_type == isc_info_sql_stmt_select and self._is_open):
+            self.trans.connection._op_free_statement(self.handle, DSQL_close)
+            if self.trans.connection.accept_type != ptype_lazy_send:
+                (h, oid, buf) = self.trans.connection._op_response()
+        self._is_open = False
 
     def drop(self):
-        if self.handle == 0 or self.handle == -1:
-            return
-        self.trans.connection._op_free_statement(self.handle, DSQL_drop)
-        if self.trans.connection.accept_type != ptype_lazy_send:
-            (h, oid, buf) = self.trans.connection._op_response()
-        self.handle = -1
-
-    def clear_handle(self):
+        DEBUG_OUTPUT("Statement::drop()")
+        if self.handle != -1:
+            self.trans.connection._op_free_statement(self.handle, DSQL_drop)
+            if self.trans.connection.accept_type != ptype_lazy_send:
+                (h, oid, buf) = self.trans.connection._op_response()
         self._is_open = False
-        self.handle = 0
+        self.handle = -1
 
     @property
     def is_opened(self):
-        return self._is_open and self.handle != 0
+        return self._is_open and self.handle != -1
 
 class PreparedStatement(object):
     def __init__(self, cur, sql, explain_plan=False):
+        DEBUG_OUTPUT("PreparedStatement::__init__()")
         self.cur = cur
         self.sql = sql
         transaction = self.cur.transaction
@@ -176,7 +173,7 @@ class PreparedStatement(object):
 
 
 def _fetch_generator(stmt):
-    stmt.open()
+    DEBUG_OUTPUT("_fetch_generator()", stmt.trans._trans_handle)
     connection = stmt.trans.connection
     more_data = True
     while more_data:
@@ -194,7 +191,7 @@ def _fetch_generator(stmt):
                         continue
                     connection._op_open_blob(r[i], stmt.trans.trans_handle)
                     (h, oid, buf) = connection._op_response()
-                    v = bytes([])
+                    v = bs([])
                     n = 1   # 1:mora data 2:no more data
                     while n == 1:
                         connection._op_get_segment(h)
@@ -208,12 +205,17 @@ def _fetch_generator(stmt):
                         (h, oid, buf) = connection._op_response()
                     r[i] = v
                     if x.sqlsubtype == 1:    # TEXT
-                        r[i] = connection.bytes_to_str(r[i])
+                        if connection.use_unicode:
+                            r[i] = connection.bytes_to_ustr(r[i])
+                        else:
+                            r[i] = connection.bytes_to_str(r[i])
             yield r
+    DEBUG_OUTPUT("_fetch_generator() StopIteration()")
     raise StopIteration()
 
 class Cursor(object):
     def __init__(self, trans):
+        DEBUG_OUTPUT("Cursor::__init__()")
         self._transaction = trans
         self.stmt = None
         self.arraysize = 1
@@ -251,27 +253,17 @@ class Cursor(object):
             stmt.prepare(query)
         return stmt
 
-    def _execute(self, stmt_handle, params):
-        cooked_params = self._convert_params(params)
-        self.transaction.connection._op_execute(stmt_handle,
-                                self.transaction.trans_handle, cooked_params)
-        try:
-            (h, oid, buf) = self.transaction.connection._op_response()
-        except OperationalError:
-            e = sys.exc_info()[1]
-            if 335544665 in e.gds_codes:
-                raise IntegrityError(e._message, e.gds_codes, e.sql_code)
-            raise OperationalError(e._message, e.gds_codes, e.sql_code)
-
     def prep(self, query, explain_plan=False):
+        DEBUG_OUTPUT("Cursor::prep()")
         prepared_statement = PreparedStatement(self, query,
                                                 explain_plan=explain_plan)
         return prepared_statement
 
     def execute(self, query, params=[]):
+        DEBUG_OUTPUT("Cursor::execute()", query, params)
         stmt = self._get_stmt(query)
+        cooked_params = self._convert_params(params)
         if stmt.stmt_type == isc_info_sql_stmt_exec_procedure:
-            cooked_params = self._convert_params(params)
             self.transaction.connection._op_execute2(stmt.handle,
                 self.transaction.trans_handle, cooked_params,
                 calc_blr(stmt.xsqlda))
@@ -280,14 +272,32 @@ class Cursor(object):
             self.transaction.connection._op_response()
             self._fetch_records = None
         else:
-            self._execute(stmt.handle, params)
+            self.transaction.connection._op_execute(stmt.handle,
+                                self.transaction.trans_handle, cooked_params)
+            try:
+                (h, oid, buf) = self.transaction.connection._op_response()
+            except OperationalError as e:
+                self._fetch_records = None
+                self._callproc_result = None
+                if e.gds_codes.intersection(
+                    [335544665, 335544466, 335544838, 335544347]):
+                    raise IntegrityError(e._message, e.gds_codes, e.sql_code)
+                if e.sql_code == -303:
+                    warnings.warn(e._message)
+                    return
+                else:
+                    raise OperationalError(e._message, e.gds_codes, e.sql_code)
+
             if stmt.stmt_type == isc_info_sql_stmt_select:
                 self._fetch_records = _fetch_generator(stmt)
+                stmt.open()
             else:
                 self._fetch_records = None
             self._callproc_result = None
+        self.transaction.is_dirty = True
 
     def callproc(self, procname, params=[]):
+        DEBUG_OUTPUT("Cursor::callproc()")
         query = 'EXECUTE PROCEDURE ' + procname + ' ' + ','.join('?'*len(params))
         return self.execute(query, params)
 
@@ -297,6 +307,8 @@ class Cursor(object):
 
 
     def fetchone(self):
+        if not self.transaction.is_dirty:
+            return None
         # callproc or not select statement
         if not self._fetch_records:
             if self._callproc_result:
@@ -327,12 +339,14 @@ class Cursor(object):
 
     def fetchall(self):
         # callproc or not select statement
+        if not self.transaction.is_dirty:
+            return None
         if not self._fetch_records:
             if self._callproc_result:
                 r = [tuple(self._callproc_result)]
                 self._callproc_result = None
                 return r
-            return None
+            return []
         # select statement
         return [tuple(r) for r in self._fetch_records]
 
@@ -345,7 +359,7 @@ class Cursor(object):
                 r = [self._callproc_result]
                 self._callproc_result = None
                 return r
-            return None
+            return []
         # select statement
         return list(itertools.islice(self._fetch_records, size))
 
@@ -371,6 +385,7 @@ class Cursor(object):
             r = self.fetchonemap()
 
     def close(self):
+        DEBUG_OUTPUT("Cursor::close()")
         if not self.stmt:
             return
         self.stmt.drop()
@@ -395,29 +410,31 @@ class Cursor(object):
 
     @property
     def rowcount(self):
-        if not self.stmt.is_opened:
+        DEBUG_OUTPUT("Cursor::rowcount()")
+        if self.stmt.handle == -1:
             return -1
 
         self.transaction.connection._op_info_sql(self.stmt.handle,
-                                     bytes([isc_info_sql_records]))
+                                     bs([isc_info_sql_records]))
         (h, oid, buf) = self.transaction.connection._op_response()
-        assert buf[:3] == bytes([0x17,0x1d,0x00]) # isc_info_sql_records
+        assert buf[:3] == bs([0x17,0x1d,0x00]) # isc_info_sql_records
         if self.stmt.stmt_type == isc_info_sql_stmt_select:
-            assert buf[17:20] == bytes([0x0d,0x04,0x00])
+            assert buf[17:20] == bs([0x0d,0x04,0x00])
                                               # isc_info_req_select_count
             count = bytes_to_int(buf[20:24])
         elif self.stmt.stmt_type == isc_info_sql_stmt_insert:
-            assert buf[24:27] == bytes([0x0e,0x04,0x00])
+            assert buf[24:27] == bs([0x0e,0x04,0x00])
                                               # isc_info_req_insert_count
             count = bytes_to_int(buf[27:31])
         elif self.stmt.stmt_type == isc_info_sql_stmt_update:
-            assert buf[3:6] == bytes([0x0f,0x04,0x00])
+            assert buf[3:6] == bs([0x0f,0x04,0x00])
                                               # isc_info_req_update_count
             count = bytes_to_int(buf[6:10])
         elif self.stmt.stmt_type == isc_info_sql_stmt_delete:
-            assert buf[10:13] == bytes([0x10,0x04,0x00])
+            assert buf[10:13] == bs([0x10,0x04,0x00])
                                               # isc_info_req_delete_count
             count = bytes_to_int(buf[13:17])
+        DEBUG_OUTPUT("Cursor::rowcount()", self.stmt.stmt_type, count)
         return count
 
 class EventConduit(WireProtocol):
@@ -465,18 +482,21 @@ class EventConduit(WireProtocol):
 
 class Connection(WireProtocol):
     def cursor(self):
+        DEBUG_OUTPUT("Connection::cursor()")
         if self._transaction is None:
             self.begin()
         return Cursor(self._transaction)
 
     def begin(self):
+        DEBUG_OUTPUT("Connection::begin()")
         if not self.sock:
             raise InternalError
         if self._transaction is None:
-            self._transaction = Transaction(self)
+            self._transaction = Transaction(self, self._autocommit)
         self._transaction.begin()
 
     def commit(self, retaining=False):
+        DEBUG_OUTPUT("Connection::commit()")
         if self._transaction:
             self._transaction.commit(retaining=retaining)
 
@@ -484,17 +504,18 @@ class Connection(WireProtocol):
         return self._transaction.savepoint(name)
 
     def rollback(self, retaining=False, savepoint=None):
+        DEBUG_OUTPUT("Connection::rollback()")
         if self._transaction:
             self._transaction.rollback(retaining=retaining, savepoint=savepoint)
 
     def execute_immediate(self, query):
         if self._transaction is None:
-            self._transaction = Transaction(self)
+            self._transaction = Transaction(self, self._autocommit)
             self._transaction.begin()
         self._op_exec_immediate(
             self._transaction.trans_handle, query=query)
         (h, oid, buf) = self._op_response()
-        self.is_dirty = True
+        self._transaction.is_dirty=True
 
     def __init__(self, dsn=None, user=None, password=None, role=None, host=None,
                     database=None, charset=DEFAULT_CHARSET, port=3050,
@@ -502,6 +523,7 @@ class Connection(WireProtocol):
                     timeout=None, isolation_level=None, use_unicode=None,
                     auth_plugin_list=('Srp', 'Legacy_Auth'),
                     wire_crypt=True, create_new=False):
+        DEBUG_OUTPUT("Connection::__init__()")
         self.db_handle = None
         if dsn:
             i = dsn.find(':')
@@ -539,16 +561,16 @@ class Connection(WireProtocol):
         self.use_unicode = use_unicode
         self.last_event_id = 0
 
+        self._autocommit = False
         self._transaction = None
         self.sock = SocketStream(self.hostname, self.port, self.timeout, cloexec)
 
         self._op_connect(auth_plugin_list, wire_crypt)
         try:
             self._op_accept()
-        except OperationalError:
+        except OperationalError as e:
             self.sock.close()
             self.sock = None
-            e = sys.exc_info()[1]
             raise e
         if create_new:                      # create database
             self._op_create(self.page_size)
@@ -568,12 +590,18 @@ class Connection(WireProtocol):
     def set_isolation_level(self, isolation_level):
         self.isolation_level = int(isolation_level)
 
+    def set_autocommit(self, is_autocommit):
+        if self._autocommit != is_autocommit and self._transaction is not None:
+            self.rollback()
+            self._transaction = None
+        self._autocommit = is_autocommit
+
     def _db_info(self, info_requests):
         if info_requests[-1] == isc_info_end:
-            self._op_info_database(bytes(info_requests))
+            self._op_info_database(bs(info_requests))
         else:
             self._op_info_database(
-                        bytes(info_requests+type(info_requests)([isc_info_end])))
+                        bs(info_requests+type(info_requests)([isc_info_end])))
         (h, oid, buf) = self._op_response()
         i = 0
         i_request = 0
@@ -685,6 +713,7 @@ class Connection(WireProtocol):
             return v
 
     def db_info(self, info_requests):
+        DEBUG_OUTPUT("Connection::db_info()")
         if type(info_requests) == int:  # singleton
             r = self._db_info([info_requests])
             return self._db_info_convert_type(info_requests, r[0][1])
@@ -705,6 +734,7 @@ class Connection(WireProtocol):
         return {}
 
     def close(self):
+        DEBUG_OUTPUT("Connection::close()")
         if self.sock is None:
             return
         if self.db_handle:
@@ -718,6 +748,7 @@ class Connection(WireProtocol):
         self.db_handle = None
 
     def drop_database(self):
+        DEBUG_OUTPUT("Connection::drop_database()")
         self._op_drop_database()
         (h, oid, buf) = self._op_response()
         self.sock.close()
@@ -735,18 +766,24 @@ class Connection(WireProtocol):
         return self.sock is None
 
 class Transaction(object):
-    def __init__(self, connection, tpb=None):
+    def __init__(self, connection, is_autocommit=False):
+        DEBUG_OUTPUT("Transaction::__init__()")
         self._connection = connection
         self._trans_handle = None
-        self.stmts = set()
+        self._autocommit = is_autocommit
 
     def _begin(self):
-        self.connection._op_transaction(
-                transaction_parameter_block[self.connection.isolation_level])
+        DEBUG_OUTPUT("Transaction::_begin()")
+        tpb = transaction_parameter_block[self.connection.isolation_level]
+        if self._autocommit:
+            tpb += bs([isc_tpb_autocommit])
+        self.connection._op_transaction(tpb)
         (h, oid, buf) = self.connection._op_response()
         self._trans_handle = h
+        self.is_dirty = False
 
     def begin(self):
+        DEBUG_OUTPUT("Transaction::begin()")
         self._begin()
 
     def savepoint(self, name):
@@ -756,23 +793,25 @@ class Transaction(object):
                         query='SAVEPOINT '+name)
         (h, oid, buf) = self.connection._op_response()
 
-    def _clear_stmts(self):
-        for s in self.stmts:
-            s.clear_handle()
-
     def commit(self, retaining=False):
+        DEBUG_OUTPUT("Transaction::commit()",
+                        self._trans_handle, retaining)
         if self._trans_handle is None:
             return
+        if not self.is_dirty:
+            return
         if retaining:
             self.connection._op_commit_retaining(self._trans_handle)
             (h, oid, buf) = self.connection._op_response()
         else:
-            self._clear_stmts()
             self.connection._op_commit(self._trans_handle)
             (h, oid, buf) = self.connection._op_response()
             self._trans_handle = None
+        self.is_dirty = False
 
     def rollback(self, retaining=False, savepoint=None):
+        DEBUG_OUTPUT("Transaction::rollback()",
+                        self._trans_handle, retaining, savepoint)
         if self._trans_handle is None:
             return
         if savepoint:
@@ -780,22 +819,24 @@ class Transaction(object):
                         query='ROLLBACK TO '+savepoint)
             (h, oid, buf) = self.connection._op_response()
             return
+        if not self.is_dirty:
+            return
         if retaining:
             self.connection._op_rollback_retaining(self._trans_handle)
             (h, oid, buf) = self.connection._op_response()
         else:
-            self._clear_stmts()
             self.connection._op_rollback(self._trans_handle)
             (h, oid, buf) = self.connection._op_response()
             self._trans_handle = None
+        self.is_dirty = False
 
     def _trans_info(self, info_requests):
         if info_requests[-1] == isc_info_end:
             self.connection._op_info_transaction(self.trans_handle,
-                    bytes(info_requests))
+                    bs(info_requests))
         else:
             self.connection._op_info_transaction(self.trans_handle,
-                    bytes(info_requests+type(info_requests)([isc_info_end])))
+                    bs(info_requests+type(info_requests)([isc_info_end])))
         (h, oid, buf) = self.connection._op_response()
         i = 0
         i_request = 0
@@ -889,12 +930,3 @@ class RowMapping(Mapping):
         return ("<RowMapping at 0x%08x with fields: %s>" %
                 (id(self), ", ".join(values)))
 
-    if not HAS_MAPPING:
-        def keys(self):
-            return list(self)
-
-        def __setitem__(self, key, value):
-            raise NotImplementedError
-
-        def __delitem__(self, key):
-            raise NotImplementedError
